(***************************************************************************
 *
 * Solution Tree Interface
 *
 * Author: Gerald Baumgartner
 * Date:   8 March 2002
 *
 ***************************************************************************)

signature SOL_TREE =
  sig
    eqtype iSet
    eqtype iSetSeq
    eqtype expTree
    eqtype sol

    datatype solTree = Leaf of sol
		     | Extended of sol * iSetSeq * solTree
		     | Reduced of sol * iSet * solTree
		     | Unary of sol * solTree
		     | Binary of sol * solTree * solTree

    eqtype solTreeSet

    exception NoSolution

    val getSoln : solTree -> sol
    val pruneSolns : solTreeSet -> solTreeSet
    val reduceSolns : solTreeSet * iSet -> solTreeSet
    val extendSolns : solTreeSet * iSet * bool -> solTreeSet
    val makeLeafSolns : expTree * iSet -> solTreeSet
    val makeUnarySolns : solTreeSet * expTree * iSet -> solTreeSet
    val makeBinarySolns : solTreeSet * solTreeSet * expTree * iSet-> solTreeSet
    val minMemFusionSet : expTree * iSet -> solTreeSet
    val minMemFusion : expTree -> solTree
    val spaceTimeTradeoffs : expTree -> solTree list
    val toString : solTree -> string
  end
