(***************************************************************************
 *
 * Tests for Expression Trees
 *
 * Author: Gerald Baumgartner
 * Date:   29 July 2001
 *
 ***************************************************************************)

structure TestExpTree =
  struct
    open ExpTree

    val A  = Fun ("A", [("i",500), ("j",100)])
    val f1 = Sum (["i"], A)
    val B  = Fun ("B", [("j",100),("k",40),("l",15)])
    val C  = Fun ("C", [("k",40), ("l",15)])
    val f2 = BinOp (Mul, B, C)
    val f3 = Sum (["l"], f2)
    val f4 = BinOp (Mul, f1, f3)
    val f5 = Sum (["j"], f4)

    val A'  = Array ("A", [("i",500), ("j",100)])
    val f1' = Sum (["i"], A')
    val B'  = Array ("B", [("j",100),("k",40),("l",15)])
    val C'  = Array ("C", [("k",40), ("l",15)])
    val f2' = BinOp (Mul, B', C')
    val f3' = Sum (["l"], f2')
    val f4' = BinOp (Mul, f1', f3')
    val f5' = Sum (["j"], f4')

    val testFusible =
	fusible A = ["i", "j"]
	andalso fusible B = ["j", "k", "l"]
	andalso fusible C = ["k", "l"]
	andalso fusible f1 = ["j"]
	andalso fusible f2 = ["j", "k", "l"]
	andalso fusible f3 = ["j", "k"]
	andalso fusible f4 = ["j", "k"]
	andalso fusible f5 = ["k"]

    val testIndices =
	indices A = ["i", "j"]
	andalso indices B = ["j", "k", "l"]
	andalso indices C = ["k", "l"]
	andalso indices f1 = ["i", "j"]
	andalso indices f2 = ["j", "k", "l"]
	andalso indices f3 = ["j", "k", "l"]
	andalso indices f4 = ["j", "k"]
	andalso indices f5 = ["j", "k"]

    val testFusedSize =
	fusedSize (B', [["i", "j"]]) = (IntInf.fromInt 0)
	andalso fusedSize (B, [["j", "k", "l"]]) = (IntInf.fromInt 1)
	andalso fusedSize (B, [["k", "l"], ["j"]]) = (IntInf.fromInt 1)
	andalso fusedSize (B, [["j", "k"]]) = (IntInf.fromInt 15)
	andalso fusedSize (B, [["j", "l"]]) = (IntInf.fromInt 40)
	andalso fusedSize (B, [["k", "l"]]) = (IntInf.fromInt 100)
	andalso fusedSize (B, [["j"]]) = (IntInf.fromInt 600)
	andalso fusedSize (B, [["k"]]) = (IntInf.fromInt 1500)
	andalso fusedSize (B, [["l"]]) = (IntInf.fromInt 4000)
	andalso fusedSize (B, []) = (IntInf.fromInt 60000)

    val testToString =
	toString f5 = concat ["sum(j)\n",
			      "  * sum(i)\n",
			      "      A(i:500, j:100)\n",
			      "    sum(l)\n",
			      "      * B(j:100, k:40, l:15)\n",
			      "        C(k:40, l:15)\n"]

    val testToExp =
	toExp f5' = "sum_j (sum_i (A[i,j]) * sum_l (B[j,k,l] * C[k,l]))"
	andalso toExp f5 = "sum_j (sum_i (A(i,j)) * sum_l (B(j,k,l) * C(k,l)))"

    val ok =
	testFusible
	andalso testIndices
	andalso testFusedSize
	andalso testToString
	andalso testToExp
  end
