(***************************************************************************
 *
 * Tests for the Fusion Tree Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   18 September 2001
 * Date:   18 March 2010
 *
 ***************************************************************************)

structure TestFusionTree =
  struct
    open ExpTree
    open FusionTree

    structure SolElem =
      struct
	type t = Solution.sol
	val eq = op=
	val toString = Solution.toString
      end

    structure SolSet = Set (structure Elem = SolElem)

    val A  = Fun ("A", [("i",500), ("j",100)])
    val f1 = Sum (["i"], A)
    val B  = Fun ("B", [("j",100),("k",40),("l",15)])
    val C  = Fun ("C", [("k",40), ("l",15)])
    val f2 = BinOp (Mul, B, C)
    val f3 = Sum (["l"], f2)
    val f4 = BinOp (Mul, f1, f3)
    val f5 = Sum (["j"], f4)

    val f5_fus = toString (buildFusionTree (SolTree.minMemFusion f5))
    val f5_fus_ok =
	f5_fus =
	  concat["(<>, <k, j>), (156,40,0), ",
		 "sum_j (sum_i (A(i,j)) * sum_l (B(j,k,l) * C(k,l)))\n",
		 "  (<k, j>, <>), (117,1,0), ",
		 "sum_i (A(i,j)) * sum_l (B(j,k,l) * C(k,l))\n",
		 "    (<>, <ij>), (100,100,0), sum_i (A(i,j))\n",
		 "      (<ij>, <>), (1,1,0), A(i,j)\n",
		 "    (<k, j>, <l>), (16,1,0), sum_l (B(j,k,l) * C(k,l))\n",
		 "      (<k, j, l>, <>), (16,1,0), B(j,k,l) * C(k,l)\n",
		 "        (<k, j, l>, <>), (1,1,0), B(j,k,l)\n",
		 "        (<k>, <l>), (15,15,0), C(k,l)\n"]

    val ok1 = f5_fus_ok andalso f5_fus = toString (expTreeToFusionTree f5)

    val A'  = Array ("A", [("i",500), ("j",100)])
    val f1' = Sum (["i"], A')
    val B'  = Array ("B", [("j",100),("k",40),("l",15)])
    val C'  = Array ("C", [("k",40), ("l",15)])
    val f2' = BinOp (Mul, B', C')
    val f3' = Sum (["l"], f2')
    val f4' = BinOp (Mul, f1', f3')
    val f5' = Sum (["j"], f4')

    val f5'_fus = toString (buildFusionTree (SolTree.minMemFusion f5'))
    val f5'_fus_ok =
	f5'_fus =
	  concat["(<>, <j, k>), (42,40,0), ",
		 "sum_j (sum_i (A[i,j]) * sum_l (B[j,k,l] * C[k,l]))\n",
		 "  (<j, k>, <>), (3,1,0), ",
		 "sum_i (A[i,j]) * sum_l (B[j,k,l] * C[k,l])\n",
		 "    (<j>, <i>), (1,1,0), sum_i (A[i,j])\n",
		 "      (<j, i>, <>), (0,0,0), A[i,j]\n",
		 "    (<j, k>, <l>), (1,1,0), sum_l (B[j,k,l] * C[k,l])\n",
		 "      (<j, k, l>, <>), (1,1,0), B[j,k,l] * C[k,l]\n",
		 "        (<j, k, l>, <>), (0,0,0), B[j,k,l]\n",
		 "        (<j, k, l>, <>), (0,0,0), C[k,l]\n"]

    val ok2 = f5'_fus_ok andalso f5'_fus = toString (expTreeToFusionTree f5')

    val T = Array ("T", [("a",30), ("c",30), ("i",4), ("k",4)])
    val N = Array ("N", [("c",30), ("d",30), ("e",7), ("l",4)])
    val D = Array ("D", [("b",30), ("e",7), ("f",7), ("l",4)])
    val S = Array ("S", [("d",30), ("j",4), ("f",7), ("k",4)])
    val m1 = BinOp (Mul, N, D)
    val m2 = Sum (["e", "l"], m1)
    val m3 = BinOp (Mul, S, m2)
    val m4 = Sum (["d", "f"], m3)
    val m5 = BinOp (Mul, T, m4)
    val m6 = Sum (["c", "k"], m5)

    val m6_fus = toString (buildFusionTree (SolTree.minMemFusion m6))
    val m6_fus_ok =
	m6_fus =
	  concat[
		 "(<>, <bc, aijk>), (14417,14400,0), sum_ck (T[a,c,i,k] * ",
		 "sum_df (S[d,j,f,k] * sum_el (N[c,d,e,l] * D[b,e,f,l])))\n",
		 "  (<bc, aijk>, <>), (18,1,0), ",
			"T[a,c,i,k] * sum_df (S[d,j,f,k] * ",
			"sum_el (N[c,d,e,l] * D[b,e,f,l]))\n",
		 "    (<bc, aijk>, <>), (0,0,0), T[a,c,i,k]\n",
		 "    (<bc>, <df, jk>), (17,16,0), sum_df (S[d,j,f,k] * ",
		 "sum_el (N[c,d,e,l] * D[b,e,f,l]))\n",
		 "      (<bc, df, jk>, <>), (2,1,0), S[d,j,f,k] * ",
		 "sum_el (N[c,d,e,l] * D[b,e,f,l])\n",
		 "        (<bc, df, jk>, <>), (0,0,0), S[d,j,f,k]\n",
		 "        (<bc, df>, <el>), (1,1,0), ",
		 "sum_el (N[c,d,e,l] * D[b,e,f,l])\n",
		 "          (<bc, df, el>, <>), (1,1,0), ",
			"N[c,d,e,l] * D[b,e,f,l]\n",
		 "            (<bc, df, el>, <>), (0,0,0), N[c,d,e,l]\n",
		 "            (<bc, df, el>, <>), (0,0,0), D[b,e,f,l]\n"]

    val ok3 = m6_fus_ok andalso m6_fus = toString (expTreeToFusionTree m6)

    val ok = ok1 andalso ok2 andalso ok3

  end

