(***************************************************************************
 *
 * Tests for Set Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   3 June 2001
 *
 ***************************************************************************)

structure TestSet =
  struct
    structure IntElem : ELEM =
      struct
	type t = Int.int
	val eq = op=
	val toString = Int.toString
      end

    structure IntSet : SET = Set (structure Elem = IntElem)

    structure IntSetElem : ELEM =
      struct
	type t = IntSet.t IntSet.set
	val eq = IntSet.eq
	val toString = IntSet.toString
      end

    structure IntSetSet : SET = Set (structure Elem = IntSetElem)

    val testEmpty =
	IntSet.empty = nil
    val testIsEmpty =
	IntSet.isEmpty nil = true
	andalso IntSet.isEmpty [1] = false
	andalso IntSet.isEmpty [1, 2, 3] = false
    val testElem =
	IntSet.elem (2, nil) = false
	andalso IntSet.elem (2, [1]) = false
	andalso IntSet.elem (2, [2]) = true
	andalso IntSet.elem (1, [1, 2, 3]) = true
	andalso IntSet.elem (2, [1, 2, 3]) = true
	andalso IntSet.elem (3, [1, 2, 3]) = true
	andalso IntSet.elem (4, [1, 2, 3]) = false
    val testSubSet =
	IntSet.subset (nil, nil) = true
	andalso IntSet.subset ([1], nil) = false
	andalso IntSet.subset (nil, [1]) = true
	andalso IntSet.subset ([1], [1]) = true
	andalso IntSet.subset ([1], [1, 2]) = true
	andalso IntSet.subset (nil, [1, 2, 3]) = true
	andalso IntSet.subset ([1], [1, 2, 3]) = true
	andalso IntSet.subset ([2], [1, 2, 3]) = true
	andalso IntSet.subset ([3], [1, 2, 3]) = true
	andalso IntSet.subset ([1, 2], [1, 2, 3]) = true
	andalso IntSet.subset ([1, 3], [1, 2, 3]) = true
	andalso IntSet.subset ([2, 3], [1, 2, 3]) = true
	andalso IntSet.subset ([1, 2, 3], [1, 2, 3]) = true
	andalso IntSet.subset ([2, 4], [1, 2, 3]) = false
    val testEq =
	IntSet.eq (nil, nil) = true
	andalso IntSet.eq (nil, [1]) = false
	andalso IntSet.eq ([1], [2]) = false
	andalso IntSet.eq ([1], [1]) = true
	andalso IntSet.eq ([1, 2], [2, 1]) = true
    val testUnion =
	IntSet.eq (IntSet.union (nil, nil), nil)
	andalso IntSet.eq (IntSet.union ([1], [1]), [1])
	andalso IntSet.eq (IntSet.union ([1], [2]), [1, 2])
	andalso IntSet.eq (IntSet.union ([2], [1]), [1, 2])
	andalso IntSet.eq (IntSet.union ([1], [1, 2]), [1, 2])
	andalso IntSet.eq (IntSet.union ([1, 2], [1]), [1, 2])
	andalso IntSet.eq (IntSet.union ([1, 2], [1, 2]), [1, 2])
    val testIntersect =
	IntSet.eq (IntSet.intersect (nil, nil), nil)
	andalso IntSet.eq (IntSet.intersect ([1], [1]), [1])
	andalso IntSet.eq (IntSet.intersect ([1], [2]), nil)
	andalso IntSet.eq (IntSet.intersect ([2], [1]), nil)
	andalso IntSet.eq (IntSet.intersect ([1], [1, 2]), [1])
	andalso IntSet.eq (IntSet.intersect ([1, 2], [1]), [1])
	andalso IntSet.eq (IntSet.intersect ([1, 2], [1, 2]), [1, 2])
    val testDifference =
	IntSet.eq (IntSet.difference (nil, nil), nil)
	andalso IntSet.eq (IntSet.difference ([1], [1]), nil)
	andalso IntSet.eq (IntSet.difference ([1], [2]), [1])
	andalso IntSet.eq (IntSet.difference ([2], [1]), [2])
	andalso IntSet.eq (IntSet.difference ([1], [1, 2]), nil)
	andalso IntSet.eq (IntSet.difference ([1, 2], [1]), [2])
	andalso IntSet.eq (IntSet.difference ([1, 2], [1, 2]), nil)
    val testPowerSet =
	IntSetSet.elem (nil, IntSet.powerSet nil)
	andalso IntSetSet.elem (nil, IntSet.powerSet [1])
	andalso IntSetSet.elem ([1], IntSet.powerSet [1])
	andalso IntSetSet.elem (nil, IntSet.powerSet [1, 2])
	andalso IntSetSet.elem ([1], IntSet.powerSet [1, 2])
	andalso IntSetSet.elem ([2], IntSet.powerSet [1, 2])
	andalso IntSetSet.elem ([1, 2], IntSet.powerSet [1, 2])
	andalso IntSetSet.elem ([2, 1], IntSet.powerSet [1, 2])
    val testToString =
	IntSet.toString nil = "{}"
	andalso IntSet.toString [1] = "{1}"
	andalso IntSet.toString [1, 2, 3] = "{1, 2, 3}"
    val ok =
	testEmpty
	andalso testIsEmpty
	andalso testElem
	andalso testSubSet
	andalso testEq
	andalso testUnion
	andalso testIntersect
	andalso testDifference
	andalso testPowerSet
	andalso testToString
  end
