(***************************************************************************
 *
 * Tests for the Solution Tree Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   10 March 2002
 * Date:   18 March 2010
 *
 ***************************************************************************)

structure TestSolTree =
  struct
    open ExpTree
    open SolTree

    structure SolElem =
      struct
	type t = Solution.sol
	val eq = op=
	val toString = Solution.toString
      end

    structure SolSet = Set (structure Elem = SolElem)

    val zero = IntInf.fromInt 0
    val one  = IntInf.fromInt 1

    val A  = Fun ("A", [("i",500), ("j",100)])
    val f1 = Sum (["i"], A)
    val B  = Fun ("B", [("j",100),("k",40),("l",15)])
    val C  = Fun ("C", [("k",40), ("l",15)])
    val f2 = BinOp (Mul, B, C)
    val f3 = Sum (["l"], f2)
    val f4 = BinOp (Mul, f1, f3)
    val f5 = Sum (["j"], f4)

    val A_soln = minMemFusionSet (A, nil)
    val A_soln_ok = length A_soln = 1 andalso
	SolSet.elem (Solution.S ([["i", "j"]],
				 (one,
				  ExpTree.fusedSize (A, [["i", "j"]]),
				  zero, zero),
				 [], zero, A),
		     map getSoln A_soln)

    val Ae_soln = extendSolns (A_soln, indices f1, true)
    val Ae_soln_ok = SolSet.eq (map getSoln A_soln, map getSoln Ae_soln)

    val f1_soln = makeUnarySolns (Ae_soln, f1, nil)
    val f1_soln_ok = length f1_soln = 1 andalso
	SolSet.elem (Solution.S ([["j"]],
				 (one,
				  ExpTree.fusedSize (f1, [["j"]]),
				  zero, zero),
				 [], zero, f1),
		     map getSoln f1_soln)

    val B_soln = minMemFusionSet (B, nil)
    val B_soln_ok = length B_soln = 1 andalso
	SolSet.elem (Solution.S ([["j", "k", "l"]],
				 (one,
				  ExpTree.fusedSize (B, [["j", "k", "l"]]),
				  zero, zero),
				 [], zero, B),
		     map getSoln B_soln)

    val C_soln = minMemFusionSet (C, nil)
    val C_soln_ok = length C_soln = 1 andalso
	SolSet.elem (Solution.S ([["k", "l"]],
				 (one,
				  ExpTree.fusedSize (C, [["k", "l"]]),
				  zero, zero),
				 [], zero, C),
		     map getSoln C_soln)

    val Be_soln = extendSolns (B_soln, indices f2, false)
    val Be_soln_ok = length Be_soln = 1 andalso hd Be_soln = hd B_soln

    val Ce_soln = extendSolns (C_soln, indices f2, false)
    val Ce_sol  = map getSoln Ce_soln
    val Ce_soln_ok = length Ce_soln = 4 andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 600,
				  ExpTree.fusedSize (C, []),
				  zero, zero),
				 [], zero, C),
		     Ce_sol) andalso
	SolSet.elem (Solution.S ([["k", "l"]],
				 (one,
				  ExpTree.fusedSize (C, [["k", "l"]]),
				  zero, zero),
				 [], zero, C),
		     Ce_sol) andalso
	SolSet.elem (Solution.S ([["k"]],
				 (IntInf.fromInt 15,
				  ExpTree.fusedSize (C, [["k"]]),
				  zero, zero),
				 [], zero, C),
		     Ce_sol) andalso
	SolSet.elem (Solution.S ([["l"]],
				 (IntInf.fromInt 40,
				  ExpTree.fusedSize (C, [["l"]]),
				  zero, zero),
				 [], zero, C),
		     Ce_sol)

    val f2_soln = makeBinarySolns (Be_soln, Ce_soln, f2, nil)
    val f2_sol  = map getSoln f2_soln
    val f2_soln_ok = length f2_soln = 4 andalso
	SolSet.elem (Solution.S ([["j", "k", "l"]],
				 (IntInf.fromInt 601,
				  ExpTree.fusedSize (f2, [["j", "k", "l"]]),
				  zero, zero),
				 [], zero, f2),
		     f2_sol) andalso
	SolSet.elem (Solution.S ([["k", "l"], ["j"]], 
				 (IntInf.fromInt 2,
				  ExpTree.fusedSize (f2, [["k", "l"], ["j"]]),
				  zero, zero),
				 [], zero, f2),
		     f2_sol) andalso
	SolSet.elem (Solution.S ([["k"], ["j", "l"]],
				 (IntInf.fromInt 16,
				  ExpTree.fusedSize (f2, [["k"], ["j", "l"]]),
				  zero, zero),
				 [], zero, f2),
		     f2_sol) andalso
	SolSet.elem (Solution.S ([["l"], ["j", "k"]],
				 (IntInf.fromInt 41,
				  ExpTree.fusedSize (f2, [["l"], ["j", "k"]]),
				  zero, zero),
				 [], zero, f2),
		     f2_sol)

    val f3_soln = makeUnarySolns (f2_soln, f3, nil)
    val f3_sol  = map getSoln f3_soln
    val f3_soln_ok = length f3_soln = 4 andalso
	SolSet.elem (Solution.S ([["j", "k"]],
				 (IntInf.fromInt 601,
				 ExpTree.fusedSize (f3, [["j", "k"]]),
				 zero, zero),
				 [], zero, f3),
		     f3_sol) andalso
	SolSet.elem (Solution.S ([["k"]],
				 (IntInf.fromInt 101,
				  ExpTree.fusedSize (f3, [["k"]]),
				  zero, zero),
				 [], zero, f3),
		     f3_sol) andalso
	SolSet.elem (Solution.S ([["k"], ["j"]],
				 (IntInf.fromInt 16,
				  ExpTree.fusedSize (f3, [["k"], ["j"]]),
				  zero, zero),
				 [], zero, f3),
		     f3_sol) andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 4040,
				  ExpTree.fusedSize (f3, []),
				  zero, zero),
				 [], zero, f3),
		     f3_sol)

    val f1e_soln = extendSolns (f1_soln, indices f4, false)
    val f1e_sol  = map getSoln f1e_soln
    val f1e_soln_ok = length f1e_soln = 2 andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 100,
				  ExpTree.fusedSize (f1, []),
				  zero, zero),
				 [], zero, f1),
		     f1e_sol) andalso
	SolSet.elem (Solution.S ([["j"]],
				 (one,
				  ExpTree.fusedSize (f1, [["j"]]),
				  zero, zero),
				 [], zero, f1),
		     f1e_sol)

    val f3e_soln = extendSolns (f3_soln, indices f4, false)
    val f3e_sol  = map getSoln f3e_soln
    val f3e_soln_ok = length f3e_soln = 2 andalso
	SolSet.elem (Solution.S ([["j", "k"]],
				 (IntInf.fromInt 601,
				  ExpTree.fusedSize (f3, [["j", "k"]]),
				  zero, zero),
				 [], zero, f3),
		     f3e_sol) andalso
(*	SolSet.elem (Solution.S ([["k"]],
				 (IntInf.fromInt 101,
				  ExpTree.fusedSize (f3, [["k"]]),
				  zero, zero),
				 [], zero, f3),
		     f3e_sol) andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 4040,
				  ExpTree.fusedSize (f3, []),
				  zero, zero),
				 [], zero, f3),
		     f3e_sol) andalso
*)	SolSet.elem (Solution.S ([["k"], ["j"]],
				 (IntInf.fromInt 16,
				  ExpTree.fusedSize (f3, [["k"], ["j"]]),
				  zero, zero),
				 [], zero, f3),
		     f3e_sol)

    val f4_soln = makeBinarySolns (f1e_soln, f3e_soln, f4, nil)
    val f4_sol  = map getSoln f4_soln
    val f4_soln_ok = length f4_soln = 3 andalso
	SolSet.elem (Solution.S ([["j"], ["k"]],
				 (IntInf.fromInt 603,
				  ExpTree.fusedSize (f4, [["j"], ["k"]]),
				  zero, zero),
				 [], zero, f4),
		     f4_sol) andalso
	SolSet.elem (Solution.S ([["j", "k"]],
				 (IntInf.fromInt 702,
				  ExpTree.fusedSize (f4, [["j", "k"]]),
				  zero, zero),
				 [], zero, f4),
		     f4_sol) andalso
	SolSet.elem (Solution.S ([["k"], ["j"]],
				 (IntInf.fromInt 117,
				  ExpTree.fusedSize (f4, [["k"], ["j"]]),
				  zero, zero),
				 [], zero, f4),
		     f4_sol)

    val f5_soln = makeUnarySolns (f4_soln, f5, nil)
    val f5_sol  = map getSoln f5_soln
    val f5_soln_ok = length f5_soln = 2 andalso
	SolSet.elem (Solution.S ([["k"]],
				 (IntInf.fromInt 117,
				  ExpTree.fusedSize (f5, [["k"]]),
				  zero, zero),
				 [], zero, f5),
		     f5_sol) andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 642,
				  ExpTree.fusedSize (f5, []),
				  zero, zero),
				 [], zero, f5),
		     f5_sol)

    val f5r_soln = reduceSolns (f5_soln, [])
    val f5r_soln_ok = length f5r_soln = 1 andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 156,
				  ExpTree.fusedSize (f5, []),
				  zero, zero),
				 [], zero, f5),
		     map getSoln f5r_soln)

    val mmfs_soln = minMemFusionSet (f5, nil)
    val mmfs_ok = SolSet.eq (map getSoln mmfs_soln, map getSoln f5_soln)

    val mmf = minMemFusion f5
    val mmf_ok = Solution.S ([], (IntInf.fromInt 156,
				  ExpTree.fusedSize (f5, []),
				  zero, zero),
			     [], zero, f5)
		   = getSoln mmf

    val ok1 =
	A_soln_ok
	andalso Ae_soln_ok
	andalso f1_soln_ok
	andalso B_soln_ok
	andalso C_soln_ok
	andalso Be_soln_ok
	andalso Ce_soln_ok
	andalso f2_soln_ok
	andalso f3_soln_ok
	andalso f1e_soln_ok
	andalso f3e_soln_ok
	andalso f4_soln_ok
	andalso f5_soln_ok
	andalso f5r_soln_ok
	andalso mmfs_ok
	andalso mmf_ok

    val A'  = Array ("A", [("i",500), ("j",100)])
    val f1' = Sum (["i"], A')
    val B'  = Array ("B", [("j",100),("k",40),("l",15)])
    val C'  = Array ("C", [("k",40), ("l",15)])
    val f2' = BinOp (Mul, B', C')
    val f3' = Sum (["l"], f2')
    val f4' = BinOp (Mul, f1', f3')
    val f5' = Sum (["j"], f4')

    val A'_soln = minMemFusionSet (A', nil)
    val A'_soln_ok = length A'_soln = 1 andalso
	SolSet.elem (Solution.S ([["*"]],
				 (zero,
				  ExpTree.fusedSize (A', [["*"]]),
				  zero, zero),
				 [], zero, A'),
		     map getSoln A'_soln)

    val A'e_soln = extendSolns (A'_soln, indices f1', true)
    val A'e_soln_ok = length A'e_soln = 1 andalso
	SolSet.elem (Solution.S ([["i", "j"]],
				 (zero,
				  ExpTree.fusedSize (A', [["i", "j"]]),
				  zero, zero),
				 [], zero, A'),
		     map getSoln A'e_soln)

    val f1'_soln = makeUnarySolns (A'e_soln, f1', nil)
    val f1'_soln_ok = length f1'_soln = 1 andalso
	SolSet.elem (Solution.S ([["j"]],
				 (one,
				  ExpTree.fusedSize (f1', [["j"]]),
				  zero, zero),
				 [], zero, f1'),
		     map getSoln f1'_soln)

    val B'_soln = minMemFusionSet (B', nil)
    val B'_soln_ok = length B'_soln = 1 andalso
	SolSet.elem (Solution.S ([["*"]],
				 (zero,
				  ExpTree.fusedSize (B', [["*"]]),
				  zero, zero),
				 [], zero, B'),
		     map getSoln B'_soln)

    val C'_soln = minMemFusionSet (C', nil)
    val C'_soln_ok = length C'_soln = 1 andalso
	SolSet.elem (Solution.S ([["*"]],
				 (zero,
				  ExpTree.fusedSize (C', [["*"]]),
				  zero, zero),
				 [], zero, C'),
		     map getSoln C'_soln)

    val B'e_soln = extendSolns (B'_soln, indices f2', false)
    val B'e_soln_ok = length B'e_soln = 1 andalso
	SolSet.elem (Solution.S ([["j", "k", "l"]],
				 (zero,
				  ExpTree.fusedSize (B', [["j", "k", "l"]]),
				  zero, zero),
				 [], zero, B'),
		     map getSoln B'e_soln)

    val C'e_soln = extendSolns (C'_soln, indices f2', false)
    val C'e_soln_ok = length C'e_soln = 1 andalso
	SolSet.elem (Solution.S ([["j", "k", "l"]],
				 (zero,
				  ExpTree.fusedSize (C', [["j", "k", "l"]]),
				  zero, zero),
				 [], zero, C'),
		     map getSoln C'e_soln)

    val f2'_soln = makeBinarySolns (B'e_soln, C'e_soln, f2', nil)
    val f2'_soln_ok = length f2'_soln = 1 andalso
	SolSet.elem (Solution.S ([["j", "k", "l"]],
				 (one,
				  ExpTree.fusedSize (f2', [["j", "k", "l"]]),
				  zero, zero),
				 [], zero, f2'),
		     map getSoln f2'_soln)

    val f3'_soln = makeUnarySolns (f2'_soln, f3', nil)
    val f3'_sol  = map getSoln f3'_soln
    val f3'_soln_ok = length f3'_soln = 1 andalso
	SolSet.elem (Solution.S ([["j", "k"]],
				 (one,
				  ExpTree.fusedSize (f3', [["j", "k"]]),
				  zero, zero),
				 [], zero, f3'),
		     map getSoln f3'_soln)

    val f1'e_soln = extendSolns (f1'_soln, indices f4', false)
    val f1'e_sol  = map getSoln f1'e_soln
    val f1'e_soln_ok = length f1'e_soln = 2 andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 100,
				  ExpTree.fusedSize (f1', []),
				  zero, zero),
				 [], zero, f1'),
		     f1'e_sol) andalso
	SolSet.elem (Solution.S ([["j"]],
				 (one,
				  ExpTree.fusedSize (f1', [["j"]]),
				  zero, zero),
				 [], zero, f1'),
		     f1'e_sol)

    val f3'e_soln = extendSolns (f3'_soln, indices f4', false)
    val f3'e_soln_ok =
	SolSet.eq (map getSoln f3'_soln, map getSoln f3'e_soln)

    val f4'_soln = makeBinarySolns (f1'e_soln, f3'e_soln, f4', nil)
    val f4'_sol  = map getSoln f4'_soln
    val f4'_soln_ok = length f4'_soln = 2 andalso
	SolSet.elem (Solution.S ([["j"], ["k"]],
				 (IntInf.fromInt 3,
				  ExpTree.fusedSize (f4', [["j"], ["k"]]),
				  zero, zero),
				 [], zero, f4'),
		     f4'_sol) andalso
	SolSet.elem (Solution.S ([["j", "k"]],
				 (IntInf.fromInt 102,
				  ExpTree.fusedSize (f4', [["j", "k"]]),
				  zero, zero),
				 [], zero, f4'),
		     f4'_sol)

    val f5'_soln = makeUnarySolns (f4'_soln, f5', nil)
    val f5'_sol  = map getSoln f5'_soln
    val f5'_soln_ok = length f5'_soln = 2 andalso
	SolSet.elem (Solution.S ([["k"]],
				 (IntInf.fromInt 102,
				  ExpTree.fusedSize (f5', [["k"]]),
				  zero, zero),
				 [], zero, f5'),
		     f5'_sol) andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 42,
				  ExpTree.fusedSize (f5', []),
				  zero, zero),
				 [], zero, f5'),
		     f5'_sol)

    val f5'r_soln = reduceSolns (f5'_soln, [])
    val f5'r_soln_ok = length f5'r_soln = 1 andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 42,
				  ExpTree.fusedSize (f5', []),
				  zero, zero),
				 [], zero, f5'),
		     map getSoln f5'r_soln)

    val mmfs'_soln = minMemFusionSet (f5', nil)
    val mmfs'_ok = SolSet.eq (map getSoln mmfs'_soln,
			      map getSoln f5'_soln)

    val mmf' = minMemFusion f5'
    val mmf'_ok = Solution.S ([], (IntInf.fromInt 42,
				   ExpTree.fusedSize (f5', []),
				   zero, zero),
			      [], zero, f5')
		    = getSoln mmf'

    val ok2 =
	A'_soln_ok
	andalso A'e_soln_ok
	andalso f1'_soln_ok
	andalso B'_soln_ok
	andalso C'_soln_ok
	andalso B'e_soln_ok
	andalso C'e_soln_ok
	andalso f2'_soln_ok
	andalso f3'_soln_ok
	andalso f1'e_soln_ok
	andalso f3'e_soln_ok
	andalso f4'_soln_ok
	andalso f5'_soln_ok
	andalso f5'r_soln_ok
	andalso mmfs'_ok
	andalso mmf'_ok

    val T = Array ("T", [("a",300), ("c",300), ("i",40), ("k",40)])
    val N = Array ("N", [("c",300), ("d",300), ("e",70), ("l",40)])
    val D = Array ("D", [("b",300), ("e",70), ("f",70), ("l",40)])
    val S = Array ("S", [("d",300), ("j",40), ("f",70), ("k",40)])
    val m1 = BinOp (Mul, N, D)
    val m2 = Sum (["e", "l"], m1)
    val m3 = BinOp (Mul, S, m2)
    val m4 = Sum (["d", "f"], m3)
    val m5 = BinOp (Mul, T, m4)
    val m6 = Sum (["c", "k"], m5)

    val m6r_soln = minMemFusionSet (m6, nil)
    val m6r_sol  = map getSoln m6r_soln
    val m6r_soln_ok = length m6r_soln = 4 andalso
	SolSet.elem (Solution.S ([],
				 (IntInf.fromInt 144480001,
				  ExpTree.fusedSize (m6, []),
				  zero, zero),
				 [], zero, m6),
		     m6r_sol) andalso
	SolSet.elem (Solution.S ([["b"],["a","i","j"]],
				 (IntInf.fromInt 480002,
				  ExpTree.fusedSize (m6,[["b"],["a","i","j"]]),
				  zero, zero),
				 [], zero, m6),
		     m6r_sol) andalso
	SolSet.elem (Solution.S ([["b"]],
				 (IntInf.fromInt 481601,
				  ExpTree.fusedSize (m6, [["b"]]),
				  zero, zero),
				 [], zero, m6),
		     m6r_sol) andalso
	SolSet.elem (Solution.S ([["a","b","i","j"]],
				 (IntInf.fromInt 144000002,
				  ExpTree.fusedSize (m6, [["a","b","i","j"]]),
				  zero, zero),
				 [], zero, m6),
		     m6r_sol)

    val m6rr_soln = minMemFusion m6
    val m6rr_soln_ok = Solution.S ([],
				   (IntInf.fromInt 144001601,
				    ExpTree.fusedSize (m6, []),
				    zero, zero),
				   [], zero, m6)
			 = getSoln m6rr_soln

    val ok3 = m6r_soln_ok andalso m6rr_soln_ok

    val ok = ok1 andalso ok2 andalso ok3

  end
